/*	main.cpp
*
*	Code for LEAF LED Charge Port Light.
*
*	Turns on LEDs at startup, dims after 4.5 minutes, then turns off lights
*	30 seconds later and enters low-power sleep mode.
*
*	Written by Cathy Saxton
*	robotics@idleloop.com
*/

#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include "Util.h"

/* ========================= CONSTANTS ========================= */

/* milliseconds to wait before changing light level */
//#define TESTING
#ifndef TESTING
static const ulong msOff = 300000;	// 5 minutes before light turned off
static const ulong msDim = 270000;	// 4.5 minutes before light dimmed
#else
static const ulong msOff = 15000;	// 15s
static const ulong msDim = 5000;	// 5s
#endif

/* values to trigger counter "match," turning LED off for rest of period;
   period is 16 ticks, so for example, 12 will be a 75% duty cycle */
static const uchar tickBright = 12;
static const uchar tickDim = 4;

/* =========================== TIMER =========================== */

/*
Timer/Counter 0 used for timing, LED dimming

MCU clock running at 1 MHz
scale /64 for 16 kHz T/C
16 ticks for 1 ms
refresh = 1 kHz, should be plenty fast for no flicker (well over 75 Hz)

CS -- 011: /64
WGM -- 111: fast PWM, top=OCRA
COM0B output
  00: disconnected
  10: OC0B set [LED on] at bottom, clear [LED off] at match
*/

static volatile ulong s_msTimer = 0;

/* starts counter (described above); COM0B output for bright LED */
inline void InitTimer()
{
	// start timing at 0
	s_msTimer = 0;
	TCNT0 = 0;

	OCR0A = 15;				// "top" -- 16 ticks (0-15) at 16 kHz = 1ms
	OCR0B = tickBright - 1;	// output match (0 to x-1 = x ticks)
	TCCR0A = 0x23;			// 00 10 .. 11 -- COM0B output (on OC0B pin), WGM low 2 bits
	TCCR0B = 0x0b;			// 00 .. 1 011 -- no FOC, WGM high bit, CS /64

	/* set Overflow Interrupt Enable (to count milliseconds) */
	SetBit(TIMSK, TOIE0);
}

/* change COM0B output to dim LED */
inline void DimLight()
{
	OCR0B = tickDim - 1;
}

inline void StopTimers()
{
	TCCR0A &= ~0x30;	// disable COM0B output
	TCCR0B &= ~0x07;	// clear CS bits to stop timer
	ClearBit(TIMSK, TOIE0);	// no more ms timer
}

/* returns number of ms since Timer initialized */
inline ulong MsCur()
{
	char sregSav;
	ulong ms;

	/* disable (delay) interrupts while we read
	   s_msTimer since it's 4 bytes and we don't
	   want it changing during the read! */
	sregSav = SREG;	// cache register holding Global Interrupt Flag
	cli();			// clear Global Interrupt Flag
	ms = s_msTimer;
	SREG = sregSav;	// restore register holding Global Interrupt Flag

	return ms;
}

/* this interrupt happens every ms, when the counter (TCNT0)
   reaches the specified OCR0A value */
ISR(TIM0_OVF_vect)
{
	++s_msTimer;
}

/* ======================= MAIN PROGRAM ======================== */

int main()
{
	/*
	*	initialization
	*/

	/* B0-4 as output; B1 is light, controlled by timer; rest are unused, set low */
	DDRB |= 0x01f;		// B0-4 as output
	PORTB &= ~0x1f;		// output low (setting B1 low irrelevant once timer takes over)

	InitTimer();	// start timer, lights on bright
	sei();			// enable interrupts

	/*
	*	wait until it's time to dim
	*/
	
	while (MsCur() < msDim)
		;

	DimLight();
	
	/*
	*	wait until it's time to turn off light
	*/
	
	while (MsCur() < msOff)
		;

	/* stop timer (and disable its output and interrupt) */
	StopTimers();
	
	/* turn off light */
	PORTB &= ~0x02;	// B1 clear/low: NPN transistor off, light off

	/*
	*	enter sleep
	*/
	
	MCUCR = 0x10;		// ...10...  SM (sleep mode) = 10 [power-down]
	cli();				// disable interrupts
	SetBit(MCUCR, SE);	// sleep enable
	sleep_cpu();
	
	/* we'd in theory get here after MCU wakes up, but nothing is enabled which will wake it! */
	/* (power cycle will of course restart the program and turn the lights on) */

	return 0;
}
